/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "../libs.h"
#include "framebuffer.h"
#include "primitive/primitive.h"
#include "shaderprogram.h"
#include "camera.h"
#include "../engine/engine.h"
#include "primitive/quad.h"
#include "../util/dynarray.h"
#include "render.h"

#define RENDER_PASSES_MAX 8

typedef struct {
  framebuffer_t frame;
  shaderprogram_t *shader;
} renderpass_t;

typedef struct {
  framebuffer_t frame;
  primitive_t quad;

  renderpass_t passes[RENDER_PASSES_MAX];
  uint8_t passCount;
} renderlist_t;

typedef struct {
  shaderprogram_t *program;
  shaderuniform_t uniformView;
  shaderuniform_t uniformPosition;
  shaderuniform_t uniformProjection;
  shaderuniform_t uniformTexture[RENDER_PASSES_MAX];
} renderlistbackshader_t;

/**
 * Initialize a render pass list.
 * 
 * @param list List to initialize.
 * @param width Width of the frame buffer(s).
 * @param height Height of the frame buffer(s).
 */
void renderListInit(renderlist_t *list, int32_t width, int32_t height);

/**
 * Retrieve the render pass at the given index.
 * 
 * @param list List to get the pass from.
 * @param pass Render pass index to get.
 * @return The render pass at the given index.
 */
renderpass_t * renderListGetPass(renderlist_t *list, uint8_t pass);

/**
 * Adds a render pass to the render list.
 * 
 * @param list Render list to add the pass to.
 * @param shader Shader to use for the render pass.
 * @return The render pass index.
 */
uint8_t renderPassAdd(renderlist_t *list, shaderprogram_t *shader);

/**
 * Prepare the rendering for a specific render pass. This will set up the 
 * render pass framebuffer, shader and prepare it for rendering.
 * 
 * @param list List to get the render pass from.
 * @param engine Game engine for the render pass.
 * @param pass Pass index to render.
 * @return The pointer to the render pass that is now active.
 */
renderpass_t * renderListRenderPass(
  renderlist_t *list, engine_t *engine, uint8_t pass
);

/**
 * Render a render list. The render list will provide the sum of the textures to
 * the given shader as uniforms and then call a single render against a textured
 * quad. The given shader will need to decide how to multiply the provided 
 * texture indexes.
 * 
 * @param list List to render.
 * @param backShader Shader to use while rendering.
 */
void renderListRender(renderlist_t *list, renderlistbackshader_t *backShader);

/**
 * Resize an existing render list and all its render pass frame buffers. This
 * will dispose (and clear) the existing frame buffers for both the list and 
 * each of the passes. Resizing to the same size won't cause an update to occur.
 * 
 * @param list List to update.
 * @param width New render list width.
 * @param height New render list height.
 */
void renderListResize(renderlist_t *list, int32_t width, int32_t height);

/**
 * Dispose a render list entirely.
 * 
 * @param list Render list to dispose.
 */
void renderListDispose(renderlist_t *list);