/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "assetmanager.h"

assetmanagerloaderdefinition_t ASSET_MANAGER_LOADERS[] = {
  {
    &_assetManagerLoaderTextureAsync,
    &_assetManagerLoaderTextureSync,
    &_assetManagerLoaderTextureDispose
  },
  {
    &_assetManagerLoaderFontAsync,
    &_assetManagerLoaderFontSync,
    &_assetManagerLoaderFontDispose
  },
  {
    &_assetManagerLoaderShaderAsync,
    &_assetManagerLoaderShaderSync,
    &_assetManagerLoaderShaderDispose
  },
  {
    &_assetManagerLoaderScaledTextureAsync,
    &_assetManagerLoaderScaledTextureSync,
    &_assetManagerLoaderScaledTextureDispose
  }
};

bool _assetManagerOnSaveManagerValueChange(
  void *usr, event_t evt, void *args[], int32_t argc
) {
  ASSERT_EQUAL(argc, 3);
  
  printf("Hello World\n");

  return true;
}

void assetManagerInit(assetmanager_t *manager, savemanager_t *save) {
  threadInit(&manager->thread, &_assetManagerThread);
  eventManagerInit(&manager->events);

  manager->thread.user = manager;
  manager->itemCount = 0;
  manager->finished = false;
  manager->holderCount = 0;
  manager->running = false;
  manager->save = save;

  manager->listenerOnSaveValueChange = eventManagerSubscribe(
    &save->events, SAVE_EVENT_VALUE_CHANGED,
    manager, &_assetManagerOnSaveManagerValueChange
  );
}

float assetManagerProgressGet(assetmanager_t *manager) {
  float done;
  uint8_t i;
  assetmanageritem_t *item;

  done = 0.0f;
  for(i = 0; i < manager->itemCount; i++) {
    item = manager->items + i;
    done += assetManagerItemIsFinished(item) ? 1 : 0;
  }
  return done / ((float)manager->itemCount);
}

float assetManagerProgressGetForHolder(
  assetmanager_t *manager, assetmanagerholder_t hold
) {
  float done;
  uint8_t i, j, c;
  assetmanageritem_t *item;
  done = 0.0f;
  c = 0x00;
  for(i = 0; i < manager->itemCount; i++) {
    //Is this held by this holder?
    item = manager->items + i;

    for(j = 0; j < item->holderCount; j++) {
      if(item->holders[j] != hold) continue;
      c++;
      done += assetManagerItemIsFinished(item) ? 1 : 0;
      break;
    }
  }
  // Is this thing even holding any assets to begin with?
  if(c == 0x00) return 1.0f;
  return done / (float)c;
}

assetmanagerholder_t assetManagerHolderCreate(assetmanager_t *man) {
  uint8_t i;

  // Find first available number.
  for(i = 0; i < 0xFF; i++) {
    if(arrayFind(
      sizeof(assetmanagerholder_t), man->holders, man->holderCount, &i
    ) == -1) break;
  }

  man->holders[man->holderCount++] = i;
  return i;// No slots left.
}

void assetManagerHolderRelease(assetmanager_t *man, assetmanagerholder_t hold) {
  int32_t i;
  uint8_t j;
  assetmanageritem_t *item;
  size_t s;

  s = sizeof(assetmanagerholder_t);
  i = arrayFind(s, man->holders,man->holderCount,&hold);
  if(i == -1) return;

  arraySplice(sizeof(uint8_t), man->holders, i, 1, man->holderCount);
  man->holderCount--;

  for(j = 0; j < man->itemCount; j++) {
    item = man->items + j;
    i = arrayFind(s, item->holders, item->holderCount, &hold);
    if(i == -1) continue;
    arraySplice(s, item->holders, i, 1, item->holderCount);
    item->holderCount--;
  }
}

void assetManagerDisposeReleased(assetmanager_t *man) {
  uint8_t i;
  assetmanagerloader_t *disp;
  assetmanageritem_t *item;

  for(i = 0; i < man->itemCount; i++) {
    item = man->items + i;
    disp = ASSET_MANAGER_LOADERS[item->type].dispose;
    if(item->holderCount > 0) continue;
    if(disp != NULL) disp(item);
    arraySplice(sizeof(assetmanageritem_t), man->items, i, 1, man->itemCount);
    man->itemCount--;
  }
}

void assetManagerDispose(assetmanager_t *man) {
  uint8_t i;
  assetmanagerloader_t *disp;
  assetmanageritem_t *item;

  eventManagerUnsubscribe(&man->save->events, man->listenerOnSaveValueChange);

  for(i = 0; i < man->itemCount; i++) {
    item = man->items + i;
    disp = ASSET_MANAGER_LOADERS[item->type].dispose;
    if(item->holderCount > 0) continue;
    if(disp != NULL) disp(item);
  }

  man->itemCount = 0;
  man->holderCount = 0;
}

// Thread Management
void assetManagerStart(assetmanager_t *manager) {
  if(manager->running) return;
  manager->running = true;
  threadStart(&manager->thread);
}

int32_t _assetManagerThread(thread_t *thread) {
  // TODO: Can I allow multiple threads to run?
  uint8_t i;
  assetmanager_t *manager;
  assetmanageritem_t *item;
  assetmanagerloaderdefinition_t *definition;
  bool result;
  manager = thread->user;

  for(i = 0; i < manager->itemCount; i++) {
    item = manager->items + i;
    definition = ASSET_MANAGER_LOADERS + item->type;

    // Only bother with ASYNC items
    if(definition->loadAsync == NULL) continue;

    // Are we already loading, not ready, or already tried to load?
    if(item->state != ASSET_MANAGER_STATE_PENDING) continue;

    // Begin loading
    item->state = ASSET_MANAGER_STATE_ASYNC_LOADING;
    result = definition->loadAsync(item);

    // Finish Loading
    if(!result) {
      item->state = ASSET_MANAGER_STATE_ASYNC_ERROR;
    } else {
      item->state = ASSET_MANAGER_STATE_ASYNC_DONE;
    }
  }

  manager->finished = assetManagerProgressGet(manager) >= 1.0f;
  manager->running = false;

  return 0;
}

void assetManagerUpdate(assetmanager_t *manager) {
  uint8_t i;
  assetmanageritem_t *item;
  assetmanagerloaderdefinition_t *definition;
  void *args[2];
  bool result;

  args[0] = manager;

  // Autostart
  if(assetManagerProgressGet(manager) < 1.0f && !manager->running) {
    assetManagerStart(manager);
  }

  for(i = 0; i < manager->itemCount; i++) {
    item = manager->items + i;
    definition = ASSET_MANAGER_LOADERS + item->type;

    // Update not ready state (Synchronously)
    if(item->state == ASSET_MANAGER_STATE_NOT_READY) {
      item->state = ASSET_MANAGER_STATE_PENDING;
      continue;
    }

    // If requires ASYNC loading, then confirm it has finished loading.
    if(definition->loadAsync != NULL) {
      if(item->state != ASSET_MANAGER_STATE_ASYNC_DONE) continue;
    } else if(item->state != ASSET_MANAGER_STATE_PENDING) {
      continue;
    }
    
    if(definition->loadSync == NULL) {
      continue;
    }
    
    // Begin sync loading
    item->state = ASSET_MANAGER_STATE_SYNC_LOADING;
    result = definition->loadSync(item);

    // Finish loading, fire events
    if(!result) {
      item->state = ASSET_MANAGER_STATE_SYNC_ERROR;
    } else {
      item->state = ASSET_MANAGER_STATE_SYNC_DONE;
      
      args[1] = item;
      eventManagerTrigger(
        &manager->events, ASSET_MANAGER_EVENT_ITEM_LOADED, args, 2
      );
    }
  }

  manager->finished = assetManagerProgressGet(manager) >= 1.0f;
}